"""
Area Chart
==========
"""
# category: basic marks
import gosling as gos

data = gos.multivec(
    url="https://resgen.io/api/v1/tileset_info/?d=UvVPeLHuRDiYA3qwFlm7xQ",
    row="sample",
    column="position",
    value="peak",
    categories=["sample 1"],
)

domain = gos.GenomicDomain(chromosome="chr1", interval=[2000500, 3000500])

track = gos.Track(data).mark_area().encode(
    x=gos.X("position:G", domain=domain, axis="bottom"),
    y="peak:Q",
    size=gos.value(2),
).properties(width=725, height=180, layout="linear")

track.view(title="Basic Marks: Area", subtitle="Tutorial Examples")



"""
Hi-C Heatmap
==============
"""
# category: basic marks
import gosling as gos
 
size = 500
data = gos.matrix("https://server.gosling-lang.org/api/v1/tileset_info/?d=leung2015-hg38")
# data = gos.matrix('/path/to/dataset.cool') # local dataset
    
track = gos.Track(data).mark_bar().encode(
  x=gos.X("xs:G", axis="bottom"),
  xe="xe:G",
  y=gos.Y("ys:G", axis="left"),
  ye="ye:G",
  color=gos.Color("value:Q", range="hot", legend=True),
).properties(width=size, height=size)

track.view()



"""
Multiscale Sequence Plot
========================
"""
# category: interactive visualizations
import gosling as gos

data = gos.multivec(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=sequence-multivec",
    row="base",
    column="position",
    value="count",
    categories=["A", "T", "G", "C"]
)

text = gos.Track(data).mark_text().encode(
    x="start:G",
    xe="end:G",
    size=gos.value(24),
    color=gos.value("white"),
    text="base:N",
).visibility_lt(
    measure="width", threshold="|xe-x|", transitionPadding=30, target="mark"
).visibility_lt(
    measure="zoomLevel", threshold=10, target="track"
).transform_filter_not("count", oneOf=[0])

bar = gos.Track(data).mark_bar().encode(
    x="position:G",
    y=gos.Y("count:Q", axis="none"),
    color=gos.Color("base:N", domain=["A", "T", "G", "C"], legend=True)
)

gos.overlay(bar, text).properties(
    title="Multi-Scale Sequence Plot",
    xDomain=gos.GenomicDomain(interval=[1000000, 1000010]),
    width=725,
    height=100
)



"""
Between Links Using Pandas
==========================
"""
# category: skip
import gosling as gos
import pandas as pd

"""
Data Transform Using Pandas
"""
df = pd.read_csv(
    "https://raw.githubusercontent.com/vigsterkr/circos/master/data/5/segdup.txt", 
    sep=" ",
    header=0,
    names=["id", "chr", "start", "end"]
)

# Use chromosome names that are interpretable in gos
df.chr = df.chr.apply(lambda x: x.replace("hs", "chr"))

# Select ids that occur exact two times
df = df[df.groupby("id")["id"].transform("size") == 2]

# Long to wide (i.e., "chr, start, end" --> "first_chr, first_start, first_end, second_chr, second_start, second_end")
df["cumcnt"] = df.groupby("id").cumcount()
df = pd.DataFrame(df.pivot(index="id", columns="cumcnt")[["chr", "start", "end"]].to_records())
df = df.rename(columns={
    "('chr', 0)": "first_chr", 
    "('chr', 1)": "second_chr", 
    "('start', 0)": "first_start", 
    "('start', 1)": "second_start", 
    "('end', 0)": "first_end", 
    "('end', 1)": "second_end"
})

df_bg = df[(df.first_chr == 'chr1') | (df.second_chr == 'chr1')]
df_hl = df[(df.first_chr != 'chr1') & (df.second_chr != 'chr1')]

column_info = [
    {"chromosomeField": "first_chr", "genomicFields": ["first_start", "first_end"]}, 
    {"chromosomeField": "second_chr", "genomicFields": ["second_start", "second_end"]}
]
data_bg = df_bg.gos.csv(genomicFieldsToConvert=column_info)
data_hl = df_hl.gos.csv(genomicFieldsToConvert=column_info)

"""
Encoding
"""
def set_encoding(track):
    return track.mark_withinLink().encode(
        x=gos.X("first_start:G"),
        xe=gos.Xe("second_end:G"),
        opacity=gos.value(0.2)
    )

gos.overlay(
    set_encoding(gos.Track(data_bg)).encode(stroke=gos.value("lightgray")),
    set_encoding(gos.Track(data_hl)).encode(stroke=gos.Stroke("second_chr:N"))
).properties(
    width=600,
    height=200
)



import hashlib
import itertools
import re
from typing import Any, Callable, List, Optional

import jsonschema

from gosling.schemapi import SchemaBase

TYPECODE_MAP = {
    "quantitative": "Q",
    "genomic": "G",
    "nominal": "N",
}

INV_TYPECODE_MAP = {v: k for k, v in TYPECODE_MAP.items()}


def parse_shorthand(
    shorthand,
    parse_types=True,
):
    """
    >>> parse_shorthand('name') == {'field': 'name'}
    True
    >>> parse_shorthand('name:Q') == {'field': 'name', 'type': 'quantitative'}
    True
    >>> parse_shorthand('foo:G') == {'field': 'foo', 'type': 'genomic'}
    True
    >>> parse_shorthand('foo:N') == {'field': 'foo', 'type': 'nominal'}
    True
    """
    if not shorthand:
        return {}

    valid_typecodes = list(TYPECODE_MAP) + list(INV_TYPECODE_MAP)

    units = dict(
        field="(?P<field>.*)",
        type="(?P<type>{})".format("|".join(valid_typecodes)),
    )

    patterns = []
    patterns.extend([r"{field}"])

    if parse_types:
        patterns = list(itertools.chain(*((p + ":{type}", p) for p in patterns)))

    regexps = (
        re.compile(r"\A" + p.format(**units) + r"\Z", re.DOTALL) for p in patterns
    )

    # find matches depending on valid fields passed
    if isinstance(shorthand, dict):
        attrs = shorthand
    else:
        attrs = next(
            exp.match(shorthand).groupdict() for exp in regexps if exp.match(shorthand)
        )

    # Handle short form of the type expression
    if "type" in attrs:
        attrs["type"] = INV_TYPECODE_MAP.get(attrs["type"], attrs["type"])

    return attrs


def infer_encoding_types(args, kwargs, channels):
    """Infer typed keyword arguments for args and kwargs
    Parameters
    ----------
    args : tuple
        List of function args
    kwargs : dict
        Dict of function kwargs
    channels : module
        The module containing all altair encoding channel classes.
    Returns
    -------
    kwargs : dict
        All args and kwargs in a single dict, with keys and types
        based on the channels mapping.
    """
    # Construct a dictionary of channel type to encoding name
    # TODO: cache this somehow?
    channel_objs = (getattr(channels, name) for name in dir(channels))
    channel_objs = (
        c for c in channel_objs if isinstance(c, type) and issubclass(c, SchemaBase)
    )
    channel_to_name = {c: c._encoding_name for c in channel_objs}
    name_to_channel = {}
    for chan, name in channel_to_name.items():
        chans = name_to_channel.setdefault(name, {})
        key = "value" if chan.__name__.endswith("Value") else "field"
        chans[key] = chan

    # First use the mapping to convert args to kwargs based on their types.
    for arg in args:
        if isinstance(arg, (list, tuple)) and len(arg) > 0:
            type_ = type(arg[0])
        else:
            type_ = type(arg)

        encoding = channel_to_name.get(type_, None)
        if encoding is None:
            raise NotImplementedError("positional of type {}" "".format(type_))
        if encoding in kwargs:
            raise ValueError("encoding {} specified twice.".format(encoding))
        kwargs[encoding] = arg

    def _wrap_in_channel_class(obj, encoding):
        if isinstance(obj, SchemaBase):
            return obj

        if isinstance(obj, str):
            obj = {"shorthand": obj}

        if isinstance(obj, (list, tuple)):
            return [_wrap_in_channel_class(subobj, encoding) for subobj in obj]

        # TODO: Separate encodings for each channel type?
        # if encoding not in name_to_channel:
        #     warnings.warn("Unrecognized encoding channel '{}'".format(encoding))
        #     return obj

        classes = name_to_channel[encoding]
        cls = classes["value"] if "value" in obj else classes["field"]

        try:
            # Don't force validation here; some objects won't be valid until
            # they're created in the context of a chart.
            return cls.from_dict(obj, validate=False)
        except jsonschema.ValidationError:
            # our attempts at finding the correct class have failed
            return obj

    return {
        encoding: _wrap_in_channel_class(obj, encoding)
        for encoding, obj in kwargs.items()
    }


def _compute_data_hash(data_str: str):
    return hashlib.md5(data_str.encode()).hexdigest()


def use_signature(Obj: Any):
    """Apply call signature and documentation of Obj to the decorated method"""

    def decorate(f):
        # call-signature of f is exposed via __wrapped__.
        # we want it to mimic Obj.__init__
        f.__wrapped__ = Obj.__init__
        f._uses_signature = Obj

        # Supplement the docstring of f with information from Obj
        if Obj.__doc__:
            doclines: List[str] = Obj.__doc__.splitlines()
            if f.__doc__:
                doc = f.__doc__ + "\n".join(doclines[1:])
            else:
                doc = "\n".join(doclines)
            try:
                f.__doc__ = doc
            except AttributeError:
                # __doc__ is not modifiable for classes in Python < 3.3
                pass

        return f

    return decorate



# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.

from . import core
from gosling.schemapi import Undefined
import gosling.utils as utils


class FieldChannelMixin(object):
    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        shorthand = self._get('shorthand')
        field = self._get('field')

        if shorthand is not Undefined and field is not Undefined:
            raise ValueError("{} specifies both shorthand={} and field={}. "
                             "".format(self.__class__.__name__, shorthand, field))

        if isinstance(shorthand, (tuple, list)):
            # If given a list of shorthands, then transform it to a list of classes
            kwds = self._kwds.copy()
            kwds.pop('shorthand')
            return [self.__class__(sh, **kwds).to_dict(validate=validate, ignore=ignore, context=context)
                    for sh in shorthand]

        if shorthand is Undefined:
            parsed = {}
        elif isinstance(shorthand, str):
            parsed = utils.parse_shorthand(shorthand)
            type_required = 'type' in self._kwds
            type_in_shorthand = 'type' in parsed
            type_defined_explicitly = self._get('type') is not Undefined
            if not (type_in_shorthand or type_defined_explicitly):
                raise ValueError("{} encoding field is specified without a type; "
                                 "the type cannot be automatically inferred because "
                                 "the data is not specified as a pandas.DataFrame."
                                 "".format(shorthand))
        else:
            # Shorthand is not a string; we pass the definition to field,
            # and do not do any parsing.
            parsed = {'field': shorthand}

        # Set shorthand to Undefined, because it's not part of the base schema.
        self.shorthand = Undefined
        self._kwds.update({k: v for k, v in parsed.items()
                           if self._get(k) is Undefined})
        return super(FieldChannelMixin, self).to_dict(
            validate=validate,
            ignore=ignore,
            context=context
        )


class ValueChannelMixin(object):
    def to_dict(self, validate=True, ignore=(), context=None):
        context = context or {}
        condition = getattr(self, 'condition', Undefined)
        copy = self  # don't copy unless we need to
        if condition is not Undefined:
            if isinstance(condition, core.SchemaBase):
                pass
            elif 'field' in condition and 'type' not in condition:
                kwds = parse_shorthand(condition['field'], context.get('data', None))
                copy = self.copy(deep=['condition'])
                copy.condition.update(kwds)
        return super(ValueChannelMixin, copy).to_dict(validate=validate,
                                                      ignore=ignore,
                                                      context=context)


class Color(FieldChannelMixin, core.Color):
    """Color schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    legend : boolean
        Whether to display legend. __Default__: `false`
    range : :class:`Range`
        Determine the colors that should be bound to data value. Default properties are
        determined considering the field type.
    scale : enum('linear', 'log')

    scaleOffset : List(float)
        Whether to use offset of the domain proportionally. This is bound to brushes on the
        color legend. __Default__: `[0, 1]`
    title : string
        Title of the legend. __Default__: `undefined`
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "color"

    def __init__(self, shorthand=Undefined, domain=Undefined, field=Undefined, legend=Undefined,
                 range=Undefined, scale=Undefined, scaleOffset=Undefined, title=Undefined,
                 type=Undefined, **kwds):
        super(Color, self).__init__(shorthand=shorthand, domain=domain, field=field, legend=legend,
                                    range=range, scale=scale, scaleOffset=scaleOffset, title=title,
                                    type=type, **kwds)


class ColorValue(ValueChannelMixin, core.ChannelValue):
    """ColorValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "color"

    def __init__(self, value, **kwds):
        super(ColorValue, self).__init__(value=value, **kwds)


class Opacity(FieldChannelMixin, core.Opacity):
    """Opacity schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    range : :class:`ValueExtent`
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "opacity"

    def __init__(self, shorthand=Undefined, domain=Undefined, field=Undefined, range=Undefined,
                 type=Undefined, **kwds):
        super(Opacity, self).__init__(shorthand=shorthand, domain=domain, field=field, range=range,
                                      type=type, **kwds)


class OpacityValue(ValueChannelMixin, core.ChannelValue):
    """OpacityValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "opacity"

    def __init__(self, value, **kwds):
        super(OpacityValue, self).__init__(value=value, **kwds)


class Row(FieldChannelMixin, core.Row):
    """Row schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    clip : boolean
        Clip row when the actual y value exceeds the max value of the y scale. Used only for
        bar marks at the moment. __Default__: `true`
    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    padding : float
        Determines the size of inner white spaces on the top and bottom of individiual rows.
        __Default__: `0`
    range : :class:`ValueExtent`
        Determine the start and end position of rendering area of this track along vertical
        axis. __Default__: `[0, height]`
    type : string
        Specify the data type
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "row"

    def __init__(self, shorthand=Undefined, clip=Undefined, domain=Undefined, field=Undefined,
                 grid=Undefined, legend=Undefined, padding=Undefined, range=Undefined, type=Undefined,
                 **kwds):
        super(Row, self).__init__(shorthand=shorthand, clip=clip, domain=domain, field=field, grid=grid,
                                  legend=legend, padding=padding, range=range, type=type, **kwds)


class RowValue(ValueChannelMixin, core.ChannelValue):
    """RowValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "row"

    def __init__(self, value, **kwds):
        super(RowValue, self).__init__(value=value, **kwds)


class Size(FieldChannelMixin, core.Size):
    """Size schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    legend : boolean
        not supported: Whether to display legend. __Default__: `false`
    range : :class:`ValueExtent`
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "size"

    def __init__(self, shorthand=Undefined, domain=Undefined, field=Undefined, legend=Undefined,
                 range=Undefined, type=Undefined, **kwds):
        super(Size, self).__init__(shorthand=shorthand, domain=domain, field=field, legend=legend,
                                   range=range, type=type, **kwds)


class SizeValue(ValueChannelMixin, core.ChannelValue):
    """SizeValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "size"

    def __init__(self, value, **kwds):
        super(SizeValue, self).__init__(value=value, **kwds)


class Stroke(FieldChannelMixin, core.Stroke):
    """Stroke schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    legend : boolean
        Whether to display legend. __Default__: `false`
    range : :class:`Range`
        Ranges of visual channel values
    scaleOffset : List(float)
        Whether to use offset of the domain proportionally. This is bound to brushes on the
        color legend. __Default__: `[0, 1]`
    title : string
        Title of the legend. __Default__: `undefined`
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "stroke"

    def __init__(self, shorthand=Undefined, domain=Undefined, field=Undefined, legend=Undefined,
                 range=Undefined, scaleOffset=Undefined, title=Undefined, type=Undefined, **kwds):
        super(Stroke, self).__init__(shorthand=shorthand, domain=domain, field=field, legend=legend,
                                     range=range, scaleOffset=scaleOffset, title=title, type=type,
                                     **kwds)


class StrokeValue(ValueChannelMixin, core.ChannelValue):
    """StrokeValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "stroke"

    def __init__(self, value, **kwds):
        super(StrokeValue, self).__init__(value=value, **kwds)


class StrokeWidth(FieldChannelMixin, core.StrokeWidth):
    """StrokeWidth schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    domain : :class:`ValueExtent`
        Values of the data
    field : string
        Name of the data field
    range : :class:`ValueExtent`
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "strokeWidth"

    def __init__(self, shorthand=Undefined, domain=Undefined, field=Undefined, range=Undefined,
                 type=Undefined, **kwds):
        super(StrokeWidth, self).__init__(shorthand=shorthand, domain=domain, field=field, range=range,
                                          type=type, **kwds)


class StrokeWidthValue(ValueChannelMixin, core.ChannelValue):
    """StrokeWidthValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "strokeWidth"

    def __init__(self, value, **kwds):
        super(StrokeWidthValue, self).__init__(value=value, **kwds)


class Text(FieldChannelMixin, core.Text):
    """Text schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    domain : List(string)
        Values of the data
    field : string
        Name of the data field
    range : List(string)
        Ranges of visual channel values
    type : enum('quantitative', 'nominal')
        Specify the data type
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "text"

    def __init__(self, shorthand=Undefined, domain=Undefined, field=Undefined, range=Undefined,
                 type=Undefined, **kwds):
        super(Text, self).__init__(shorthand=shorthand, domain=domain, field=field, range=range,
                                   type=type, **kwds)


class TextValue(ValueChannelMixin, core.ChannelValue):
    """TextValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "text"

    def __init__(self, value, **kwds):
        super(TextValue, self).__init__(value=value, **kwds)


class Tooltip(FieldChannelMixin, core.Tooltip):
    """Tooltip schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    alt : string
        Name of the data field for showing in the tooltip. Will use the field name if not
        specified.
    field : string
        Specifiy a data field whose value will show in the tooltip.
    format : string
        format of the data value.
    type : :class:`FieldType`
        Type of the data field.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "tooltip"

    def __init__(self, shorthand=Undefined, alt=Undefined, field=Undefined, format=Undefined,
                 type=Undefined, **kwds):
        super(Tooltip, self).__init__(shorthand=shorthand, alt=alt, field=field, format=format,
                                      type=type, **kwds)


class X(FieldChannelMixin, core.X):
    """X schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    domain : :class:`GenomicDomain`
        Values of the data
    field : string
        Name of the data field.
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : string
        Specify the data type.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "x"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, domain=Undefined,
                 field=Undefined, grid=Undefined, legend=Undefined, linkingId=Undefined,
                 range=Undefined, type=Undefined, **kwds):
        super(X, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis, domain=domain,
                                field=field, grid=grid, legend=legend, linkingId=linkingId, range=range,
                                type=type, **kwds)


class XValue(ValueChannelMixin, core.ChannelValue):
    """XValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "x"

    def __init__(self, value, **kwds):
        super(XValue, self).__init__(value=value, **kwds)


class X1(FieldChannelMixin, core.X):
    """X1 schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    domain : :class:`GenomicDomain`
        Values of the data
    field : string
        Name of the data field.
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : string
        Specify the data type.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "x1"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, domain=Undefined,
                 field=Undefined, grid=Undefined, legend=Undefined, linkingId=Undefined,
                 range=Undefined, type=Undefined, **kwds):
        super(X1, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis, domain=domain,
                                 field=field, grid=grid, legend=legend, linkingId=linkingId,
                                 range=range, type=type, **kwds)


class X1Value(ValueChannelMixin, core.ChannelValue):
    """X1Value schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "x1"

    def __init__(self, value, **kwds):
        super(X1Value, self).__init__(value=value, **kwds)


class X1e(FieldChannelMixin, core.X):
    """X1e schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    domain : :class:`GenomicDomain`
        Values of the data
    field : string
        Name of the data field.
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : string
        Specify the data type.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "x1e"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, domain=Undefined,
                 field=Undefined, grid=Undefined, legend=Undefined, linkingId=Undefined,
                 range=Undefined, type=Undefined, **kwds):
        super(X1e, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis, domain=domain,
                                  field=field, grid=grid, legend=legend, linkingId=linkingId,
                                  range=range, type=type, **kwds)


class X1eValue(ValueChannelMixin, core.ChannelValue):
    """X1eValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "x1e"

    def __init__(self, value, **kwds):
        super(X1eValue, self).__init__(value=value, **kwds)


class Xe(FieldChannelMixin, core.X):
    """Xe schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    domain : :class:`GenomicDomain`
        Values of the data
    field : string
        Name of the data field.
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : string
        Specify the data type.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "xe"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, domain=Undefined,
                 field=Undefined, grid=Undefined, legend=Undefined, linkingId=Undefined,
                 range=Undefined, type=Undefined, **kwds):
        super(Xe, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis, domain=domain,
                                 field=field, grid=grid, legend=legend, linkingId=linkingId,
                                 range=range, type=type, **kwds)


class XeValue(ValueChannelMixin, core.ChannelValue):
    """XeValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "xe"

    def __init__(self, value, **kwds):
        super(XeValue, self).__init__(value=value, **kwds)


class Y(FieldChannelMixin, core.Y):
    """Y schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    baseline : anyOf(string, float)
        Custom baseline of the y-axis. __Default__: `0`
    domain : anyOf(:class:`ValueExtent`, :class:`GenomicDomain`)
        Values of the data
    field : string
        Name of the data field.
    flip : boolean
        Whether to flip the y-axis. This is done by inverting the `range` property.
        __Default__: `false`
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : enum('quantitative', 'nominal', 'genomic')
        Specify the data type.
    zeroBaseline : boolean
        Specify whether to use zero baseline. __Default__: `true`
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "y"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, baseline=Undefined,
                 domain=Undefined, field=Undefined, flip=Undefined, grid=Undefined, legend=Undefined,
                 linkingId=Undefined, range=Undefined, type=Undefined, zeroBaseline=Undefined, **kwds):
        super(Y, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis, baseline=baseline,
                                domain=domain, field=field, flip=flip, grid=grid, legend=legend,
                                linkingId=linkingId, range=range, type=type, zeroBaseline=zeroBaseline,
                                **kwds)


class YValue(ValueChannelMixin, core.ChannelValue):
    """YValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "y"

    def __init__(self, value, **kwds):
        super(YValue, self).__init__(value=value, **kwds)


class Y1(FieldChannelMixin, core.Y):
    """Y1 schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    baseline : anyOf(string, float)
        Custom baseline of the y-axis. __Default__: `0`
    domain : anyOf(:class:`ValueExtent`, :class:`GenomicDomain`)
        Values of the data
    field : string
        Name of the data field.
    flip : boolean
        Whether to flip the y-axis. This is done by inverting the `range` property.
        __Default__: `false`
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : enum('quantitative', 'nominal', 'genomic')
        Specify the data type.
    zeroBaseline : boolean
        Specify whether to use zero baseline. __Default__: `true`
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "y1"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, baseline=Undefined,
                 domain=Undefined, field=Undefined, flip=Undefined, grid=Undefined, legend=Undefined,
                 linkingId=Undefined, range=Undefined, type=Undefined, zeroBaseline=Undefined, **kwds):
        super(Y1, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis, baseline=baseline,
                                 domain=domain, field=field, flip=flip, grid=grid, legend=legend,
                                 linkingId=linkingId, range=range, type=type, zeroBaseline=zeroBaseline,
                                 **kwds)


class Y1Value(ValueChannelMixin, core.ChannelValue):
    """Y1Value schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "y1"

    def __init__(self, value, **kwds):
        super(Y1Value, self).__init__(value=value, **kwds)


class Y1e(FieldChannelMixin, core.Y):
    """Y1e schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    baseline : anyOf(string, float)
        Custom baseline of the y-axis. __Default__: `0`
    domain : anyOf(:class:`ValueExtent`, :class:`GenomicDomain`)
        Values of the data
    field : string
        Name of the data field.
    flip : boolean
        Whether to flip the y-axis. This is done by inverting the `range` property.
        __Default__: `false`
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : enum('quantitative', 'nominal', 'genomic')
        Specify the data type.
    zeroBaseline : boolean
        Specify whether to use zero baseline. __Default__: `true`
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "y1e"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, baseline=Undefined,
                 domain=Undefined, field=Undefined, flip=Undefined, grid=Undefined, legend=Undefined,
                 linkingId=Undefined, range=Undefined, type=Undefined, zeroBaseline=Undefined, **kwds):
        super(Y1e, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis,
                                  baseline=baseline, domain=domain, field=field, flip=flip, grid=grid,
                                  legend=legend, linkingId=linkingId, range=range, type=type,
                                  zeroBaseline=zeroBaseline, **kwds)


class Y1eValue(ValueChannelMixin, core.ChannelValue):
    """Y1eValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "y1e"

    def __init__(self, value, **kwds):
        super(Y1eValue, self).__init__(value=value, **kwds)


class Ye(FieldChannelMixin, core.Y):
    """Ye schema wrapper

    Mapping(required=[shorthand])

    Attributes
    ----------

    shorthand : string
        shorthand for field, aggregate, and type
    aggregate : :class:`Aggregate`
        Specify how to aggregate data. __Default__: `undefined`
    axis : :class:`AxisPosition`
        Specify where should the axis be put
    baseline : anyOf(string, float)
        Custom baseline of the y-axis. __Default__: `0`
    domain : anyOf(:class:`ValueExtent`, :class:`GenomicDomain`)
        Values of the data
    field : string
        Name of the data field.
    flip : boolean
        Whether to flip the y-axis. This is done by inverting the `range` property.
        __Default__: `false`
    grid : boolean
        Whether to display grid. __Default__: `false`
    legend : boolean
        Whether to display legend. __Default__: `false`
    linkingId : string
        Users need to assign a unique linkingId for [linking
        views](/docs/user-interaction#linking-views) and [Brushing and
        Linking](/docs/user-interaction#brushing-and-linking)
    range : :class:`ValueExtent`
        Values of the visual channel.
    type : enum('quantitative', 'nominal', 'genomic')
        Specify the data type.
    zeroBaseline : boolean
        Specify whether to use zero baseline. __Default__: `true`
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "ye"

    def __init__(self, shorthand=Undefined, aggregate=Undefined, axis=Undefined, baseline=Undefined,
                 domain=Undefined, field=Undefined, flip=Undefined, grid=Undefined, legend=Undefined,
                 linkingId=Undefined, range=Undefined, type=Undefined, zeroBaseline=Undefined, **kwds):
        super(Ye, self).__init__(shorthand=shorthand, aggregate=aggregate, axis=axis, baseline=baseline,
                                 domain=domain, field=field, flip=flip, grid=grid, legend=legend,
                                 linkingId=linkingId, range=range, type=type, zeroBaseline=zeroBaseline,
                                 **kwds)


class YeValue(ValueChannelMixin, core.ChannelValue):
    """YeValue schema wrapper

    Mapping(required=[value])

    Attributes
    ----------

    value : anyOf(float, string)
        Assign a constant value for a visual channel.
    """
    _class_is_valid_at_instantiation = False
    _encoding_name = "ye"

    def __init__(self, value, **kwds):
        super(YeValue, self).__init__(value=value, **kwds)



"""
Gene Annotation
===============

Re-implementation of `Higlass gene-annotations track`_

.. _Higlass gene-annotations track: https://docs.higlass.io/track_types.html#gene-annotations
"""
# category: composite visualizations
import gosling as gos

genes = gos.beddb(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=gene-annotation",
    genomicFields=[
      {"index": 1, "name": "start"},
      {"index": 2, "name": "end"}
    ],
    valueFields=[
      {"index": 5, "name": "strand", "type": "nominal"},
      {"index": 3, "name": "name", "type": "nominal"}
    ],
    exonIntervalFields=[
      {"index": 12, "name": "start"},
      {"index": 13, "name": "end"}
    ]
)

base = gos.Track(genes).encode(
    row=gos.Row("strand:N", domain=["+", "-"]),
    color=gos.Color("strand:N", domain=["+", "-"], range=["#7585FF", "#FF8A85"]),
    tooltip=["start:G", "end:G", "strand:N", "name:N"],
).properties(
    title="Genes | hg38",
)

plusGeneHead = base.mark_triangleRight(align="left").encode(
    x=gos.X("end:G", axis="top"),
    size=gos.value(15)
).transform_filter(
    field="type", oneOf=["gene"]
).transform_filter(
    field="strand", oneOf=["+"]
)

minusGeneHead = base.mark_triangleLeft(align="right").encode(
    x=gos.X("start:G", axis="top"),
    size=gos.value(15)
).transform_filter(
    field="type", oneOf=["gene"]
).transform_filter(
    field="strand", oneOf=["-"]
)

geneLabel = base.mark_text(dy=15).encode(
    x=gos.X("start:G", axis="top"),
    xe="end:G",
    text="name:N",
    size=gos.value(15)
).transform_filter(
    field="type", oneOf=["gene"]
).visibility_lt(
    measure="width",
    threshold="|xe-x|",
    transitionPadding=10,
    target="mark",
)

exon = base.mark_rect().encode(
    x=gos.X("start:G", axis="top"),
    xe="end:G",
    size=gos.value(15)
).transform_filter(
    field="type", oneOf=["exon"]
)

plusGeneRange = base.mark_rule(linePattern={"type": "triangleRight", "size": 5}).encode(
    x=gos.X("start:G", axis="top"),
    xe="end:G",
    strokeWidth=gos.value(3),
).transform_filter(
    field="type", oneOf=["gene"]
).transform_filter(
    field="strand", oneOf=["+"]
)

minusGeneRange = base.mark_rule(linePattern={"type": "triangleLeft", "size": 5}).encode(
    x=gos.X("start:G", axis="top"),
    xe="end:G",
    strokeWidth=gos.value(3),
).transform_filter(
    field="type", oneOf=["gene"]
).transform_filter(
    field="strand", oneOf=["-"]
)

gos.overlay(
    plusGeneRange, minusGeneRange, exon, plusGeneHead, minusGeneHead, geneLabel
).properties(
    width=725, height=100,
    title="Gene Annotation",
    xDomain=gos.GenomicDomain(chromosome="chr1", interval=[103400000, 103700000]),
    assembly="hg38",
    layout="linear",
    centerRadius=0.7,
)



"""
Point Plot
==========
"""
# category: basic marks
import gosling as gos

data = gos.multivec(
    url="https://resgen.io/api/v1/tileset_info/?d=UvVPeLHuRDiYA3qwFlm7xQ",
    row="sample",
    column="position",
    value="peak",
    categories=["sample 1"],
    binSize=5,
)

domain = gos.GenomicDomain(chromosome="chr1", interval=[1, 30005000])

track = gos.Track(data).mark_point().encode(
    x=gos.X("position:G", domain=domain, axis="bottom"),
    y="peak:Q",
    size="peak:Q",
    color="sample:N",
).properties(layout="linear", width=725, height=180)

track.view(title="Basic Marks: Point", subtitle="Tutorial Examples")



from __future__ import annotations

import json
import uuid
from typing import Any, Dict

import jinja2

from gosling.plugin_registry import PluginRegistry
from gosling.schema import SCHEMA_VERSION, THEMES

# TODO: Ideally we could use a single import but this seems to work ok.
HTML_TEMPLATE = jinja2.Template(
    """
<!DOCTYPE html>
<html>
<head>
  <style>.error { color: red; }</style>
</head>
<body>
  <div id="{{ output_div }}"></div>
  <script type="module">
    import { importWithMap } from "https://unpkg.com/dynamic-importmap@0.1.0";
    let importMap = {{ import_map }};
    let gosling = await importWithMap("gosling.js", importMap);
    let el = document.getElementById('{{ output_div }}');
    let spec = {{ spec }};
    let opt = {{ embed_options }};
    gosling.embed(el, spec, opt).catch((err) => {
      el.innerHTML = `\
    <div class="error">
        <p>JavaScript Error: ${error.message}</p>
        <p>This usually means there's a typo in your Gosling specification. See the javascript console for the full traceback.</p>
    </div>`;
      throw error;
    });
  </script>
</body>
</html>
"""
)

GoslingSpec = Dict[str, Any]


def get_gosling_import_map(
    gosling_version: str = SCHEMA_VERSION.lstrip("v"),
    higlass_version: str = "1.13",
    react_version: str = "18",
    pixijs_version: str = "6",
) -> dict:
    return {
        "imports": {
            "react": f"https://esm.sh/react@{react_version}",
            "react-dom": f"https://esm.sh/react-dom@{react_version}",
            "pixi": f"https://esm.sh/pixi.js@{pixijs_version}?bundle",
            "higlass": f"https://esm.sh/higlass@{higlass_version}?external=react,react-dom,pixi&bundle",
            "gosling.js": f"https://esm.sh/gosling.js@{gosling_version}?external=react,react-dom,pixi,higlass&bundle",
        }
    }


def spec_to_html(
    spec: GoslingSpec,
    output_div: str = "vis",
    embed_options: Dict[str, Any] | None = None,
    **kwargs,
):
    embed_options = embed_options or dict(padding=0, theme=themes.get())
    deps = get_gosling_import_map(**kwargs)

    return HTML_TEMPLATE.render(
        spec=json.dumps(spec),
        embed_options=json.dumps(embed_options),
        output_div=output_div,
        import_map=json.dumps(deps),
    )


class Renderer:
    def __init__(self, output_div: str = "jupyter-gosling-{}", **kwargs: Any):
        self._output_div = output_div
        self.kwargs = kwargs

    @property
    def output_div(self) -> str:
        return self._output_div.format(uuid.uuid4().hex)

    def __call__(self, spec: GoslingSpec, **meta: Any) -> Dict[str, Any]:
        raise NotImplementedError()


class HTMLRenderer(Renderer):
    def __call__(self, spec: GoslingSpec, **meta: Any):
        kwargs = self.kwargs.copy()
        kwargs.update(meta)
        html = spec_to_html(spec=spec, output_div=self.output_div, **kwargs)
        return {"text/html": html}


renderers = PluginRegistry[Renderer]("gosling.renderer")
renderers.register("default", HTMLRenderer())
renderers.enable("default")

CustomTheme = Dict[str, Any]

themes = PluginRegistry[CustomTheme]("gosling.theme")
for theme in THEMES:
    # Add builtin string themes
    themes.register(theme, theme)  # type: ignore



from __future__ import annotations

import json
import pathlib
import warnings
import typing

import jinja2
from docutils import nodes
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import flag

from gosling.display import get_gosling_import_map
from gosling.schemapi import SchemaValidationError
from gosling.utils.execeval import eval_block

if typing.TYPE_CHECKING:
    from sphinx.application import Sphinx

TEMPLATE = jinja2.Template(
    """
<div id="{{ div_id }}">
<script type="module">
  import * as gosling from "gosling.js";
  document.addEventListener("DOMContentLoaded", async () => {
      let el = document.querySelector('#{{ div_id }}');
      let spec = {{ spec }};
      let opt = { padding: 0 };
      console.log(spec);
      gosling.embed(el, spec, opt).catch(console.err);
  });
</script>
</div>
"""
)


class gosling_plot(nodes.General, nodes.Element):
    pass


class GoslingPlotDirective(Directive):
    has_content = True

    option_spec = {"code-below": flag}

    def run(self):
        env = self.state.document.settings.env
        app = env.app

        show_code = True
        code_below = "code-below" in self.options

        code = "\n".join(self.content)

        if show_code:
            source_literal = nodes.literal_block(code, code)
            source_literal["language"] = "python"

        # get the name of the source file we are currently processing
        rst_source = pathlib.Path(self.state_machine.document["source"])
        rst_dir = rst_source.parent
        rst_filename = rst_source.name

        # use the source file name to construct a friendly target_id
        serialno = env.new_serialno("gosling-plot")
        rst_base = rst_filename.replace(".", "-")
        div_id = f"{rst_base}-gosling-plot-{serialno}"
        target_id = f"{rst_base}-gosling-source-{serialno}"
        target_node = nodes.target("", "", ids=[target_id])

        # create the node in which the plot will appear;
        # this will be processed by html_visit_gosling_plot
        plot_node = gosling_plot()
        plot_node["target_id"] = target_id
        plot_node["div_id"] = div_id
        plot_node["code"] = code
        plot_node["relpath"] = str(rst_dir / env.srcdir)
        plot_node["rst_source"] = rst_source
        plot_node["rst_lineno"] = self.lineno
        plot_node["output"] = "html"

        result = [target_node]

        if code_below:
            result += [plot_node]
        if show_code:
            result += [source_literal]
        if not code_below:
            result += [plot_node]

        return result


def add_custom_head(
    app: Sphinx, pagename: str, templatename: str, context: dict, doctree
):
    custom_html = (
        f'<script type="importmap">{json.dumps(get_gosling_import_map())}</script>'
    )
    if context.get("metatags, None"):
        context["metatags"] += custom_html
    else:
        context["metatags"] = custom_html


def html_visit_gosling_plot(self, node):
    # Execute the code, saving output and namespace
    try:
        chart = eval_block(node["code"])
    except Exception as e:
        warnings.warn(
            "gosling-plot: {}:{} Code Execution failed:" "{}: {}".format(
                node["rst_source"], node["rst_lineno"], e.__class__.__name__, str(e)
            )
        )
        raise nodes.SkipNode

    # Last line should be a chart; convert to spec dict
    try:
        spec = chart.to_json()
    except SchemaValidationError:
        raise ValueError("Invalid chart: {0}".format(node["code"]))

    # Pass relevant info into the template and append to the output
    html = TEMPLATE.render(div_id=node["div_id"], spec=spec)
    self.body.append(html)


def depart_gosling_plot(self, node):
    return


def setup(app: Sphinx):
    app.connect("html-page-context", add_custom_head)
    app.add_directive("gosling-plot", GoslingPlotDirective)
    app.add_node(gosling_plot, html=(html_visit_gosling_plot, depart_gosling_plot))
    return {"version": "0.1"}



"""
Multiscale Lollipop Plot
========================
"""
# category: interactive visualizations
import gosling as gos
 
clin_var_beddb = gos.beddb(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=clinvar-beddb",
    genomicFields=[{"index": 1, "name": "start"}, {"index": 2, "name": "end"}],
    valueFields=[{"index": 7, "name": "significance", "type": "nominal"}],
)

categories = [
    "Benign", "Benign/Likely_benign", "Likely_benign", 
    "Uncertain_significance", "Likely_pathogenic", 
    "Pathogenic/Likely_pathogenic", "Pathogenic",
]

colors = ["#5A9F8C", "#5A9F8C", "#029F73", "gray", "#CB96B3", "#CB71A3", "#CB3B8C"]

clin_var_multivec = gos.multivec(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=clinvar-multivec",
    row="significance",
    column="position",
    value="count",
    categories=categories,
    binSize=4,
)


lolipop = gos.overlay(
    (
        gos.Track(clin_var_beddb).mark_bar().encode(
            x="start:G",
            y=gos.Y("significance:N", domain=categories, range=[150, 20], baseline="Uncertain_significance"),
            size=gos.value(1),
            color=gos.value("lightgray"),
            stroke=gos.value("lightgray"),
            strokeWidth=gos.value(1),
            opacity=gos.value(0.3),
        ).visibility_lt(
            measure="zoomLevel",
            target="mark",
            threshold=100000,
            transitionPadding=100000,
        )
    ),
    (
        gos.Track(clin_var_beddb).mark_point().encode(
            x="start:G",
            color=gos.Color("significance:N", domain=categories, range=colors),
            row=gos.Row("significance:N", domain=categories),
            size=gos.value(7),
            opacity=gos.value(0.8),
        ).visibility_lt(
            measure="zoomLevel",
            target="mark",
            threshold=1000000,
            transitionPadding=1000000,
        )
    ),
    (
        gos.Track(clin_var_multivec).mark_bar().encode(
            x="start:G",
            xe="end:G",
            y=gos.Y("count:Q", axis="none"),
            color=gos.Color("significance:N", domain=categories, range=colors, legend=True)
        ).visibility_gt(
            measure="zoomLevel",
            target="mark",
            threshold=500000,
            transitionPadding=500000,
        )
    ),
    width=725,
    height=150,
    xDomain=gos.GenomicDomain(chromosome="chr13", interval=[31500000, 33150000]),
)

lolipop



import pathlib


def iter_examples():
    example_dir = pathlib.Path(__file__).parent.absolute()
    for fp in example_dir.iterdir():
        if fp.name.startswith("_") or fp.suffix != ".py":
            continue
        yield fp



import ast
import dataclasses
import io
import itertools
import pathlib
import re
import tokenize
from typing import Iterable, Optional, Tuple, TypeVar

import jinja2

from gosling.examples import iter_examples

GALLERY_TEMPLATE = jinja2.Template(
    """
.. This document is auto-generated by the gosling-gallery extension. Do not modify directly.

.. _{{ gallery_ref }}:

{{ title }}
{% for char in title %}-{% endfor %}

This gallery contains a selection of examples of the visualizations **gos** can create.

{% for category, examples in example_groups %}

{{ category | capitalize }}
{% for char in category %}-{% endfor %}

.. toctree::

{% for example in examples %}
    {{ example.name }}
{%- endfor %}

{%- endfor %}
"""
)

EXAMPLE_TEMPLATE = jinja2.Template(
    """
.. This document is auto-generated by the gosling-gallery extension. Do not modify directly.

.. _gallery_{{ name }}:

{{ docstring }}

.. gosling-plot::
    {% if code_below %}:code-below:{% endif %}

    {{ code | indent(4) }}
"""
)


@dataclasses.dataclass
class Example:
    name: str
    docstring: str
    code: str
    lineno: int
    category: str

    @classmethod
    def from_file(cls, file: pathlib.Path):
        content = file.read_text()
        # change from Windows format to UNIX for uniformity
        content = content.replace("\r\n", "\n")
        docstring = ast.get_docstring(ast.parse(content)) or ""

        # Find the category comment
        find_category = re.compile(r"^#\s*category:\s*(.*)$", re.MULTILINE)
        match = find_category.search(content)
        if match is not None:
            category = match.groups()[0]
            # remove this comment from the content
            content = find_category.sub("", content)
        else:
            category = "other"

        ts = tokenize.tokenize(io.BytesIO(content.encode("utf-8")).readline)
        ds_lines = 0
        # find the first string according to the tokenizer and get
        # it's end row
        for tk in ts:
            if tk.exact_type == 3:
                ds_lines, _ = tk.end
                break
        # grab the rest of the file
        rest = "\n".join(content.split("\n")[ds_lines:])
        lineno = ds_lines + 1

        return cls(
            name=file.stem,
            docstring=docstring,
            code=rest,
            lineno=lineno,
            category=category,
        )


T = TypeVar("T")


def prev_this_next(
    it: Iterable[T], sentinel=None
) -> Iterable[Tuple[Optional[T], T, Optional[T]]]:
    """Utility to return (prev, this, next) tuples from an iterator"""
    i1, i2, i3 = itertools.tee(it, 3)
    next(i3, None)
    return zip(itertools.chain([sentinel], i1), i2, itertools.chain(i3, [sentinel]))


def populate_examples():
    return sorted(
        filter(
            lambda e: e.category != "skip",
            map(Example.from_file, iter_examples()),
        ),
        key=lambda e: e.category,
    )


def main(app):
    srcdir = pathlib.Path(app.builder.srcdir)
    title = "Example Gallery"
    gallery_dir = "gallery"
    gallery_ref = "example-gallery"
    target_dir = srcdir / gallery_dir

    if not target_dir.is_dir():
        target_dir.mkdir(parents=True)

    examples = populate_examples()
    example_groups = itertools.groupby(examples, key=lambda e: e.category)

    # Write the gallery index file
    with open(target_dir / "index.rst", "w") as f:
        f.write(
            GALLERY_TEMPLATE.render(
                title=title,
                example_groups=example_groups,
                gallery_ref=gallery_ref,
            )
        )

    for prev_ex, example, next_ex in prev_this_next(examples):
        ex = dataclasses.asdict(example)
        ex["code_below"] = True
        if prev_ex:
            ex["prev_ref"] = f"gallery_{prev_ex.name}"
        if next_ex:
            ex["next_ref"] = f"gallery_{next_ex.name}"
        with open(target_dir / (example.name + ".rst"), "w") as f:
            f.write(EXAMPLE_TEMPLATE.render(ex))


def setup(app):
    app.connect("builder-inited", main)



# The contents of this file are automatically written by
# tools/generate_schema_wrapper.py. Do not modify directly.
from . import core
from gosling.schemapi import Undefined
from typing import TypeVar


T = TypeVar('T')
class MarkMethodMixin(object):
    """A mixin class that defines mark methods"""

    def mark_point(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                   brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                   enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                   linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                   linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                   outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                   textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                   textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'point'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "point"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_line(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                  brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                  enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                  linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                  linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                  outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                  textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                  textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'line'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "line"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_area(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                  brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                  enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                  linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                  linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                  outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                  textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                  textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'area'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "area"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_bar(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                 brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                 enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                 linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                 linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                 outlineWidth=Undefined, select=Undefined, textAnchor=Undefined, textFontSize=Undefined,
                 textFontWeight=Undefined, textStroke=Undefined, textStrokeWidth=Undefined,
                 withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'bar'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "bar"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_rect(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                  brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                  enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                  linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                  linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                  outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                  textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                  textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'rect'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "rect"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_text(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                  brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                  enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                  linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                  linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                  outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                  textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                  textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'text'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "text"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_withinLink(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                        brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                        enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                        linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                        linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined,
                        outline=Undefined, outlineWidth=Undefined, select=Undefined,
                        textAnchor=Undefined, textFontSize=Undefined, textFontWeight=Undefined,
                        textStroke=Undefined, textStrokeWidth=Undefined,
                        withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'withinLink'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "withinLink"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_betweenLink(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                         brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                         enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                         linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                         linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined,
                         outline=Undefined, outlineWidth=Undefined, select=Undefined,
                         textAnchor=Undefined, textFontSize=Undefined, textFontWeight=Undefined,
                         textStroke=Undefined, textStrokeWidth=Undefined,
                         withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'betweenLink'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "betweenLink"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_rule(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                  brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                  enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                  linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                  linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                  outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                  textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                  textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'rule'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "rule"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_triangleLeft(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                          brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined,
                          dy=Undefined, enableSmoothPath=Undefined, inlineLegend=Undefined,
                          legendTitle=Undefined, linePattern=Undefined, linkConnectionType=Undefined,
                          linkMinHeight=Undefined, linkStyle=Undefined, matrixExtent=Undefined,
                          mouseOver=Undefined, outline=Undefined, outlineWidth=Undefined,
                          select=Undefined, textAnchor=Undefined, textFontSize=Undefined,
                          textFontWeight=Undefined, textStroke=Undefined, textStrokeWidth=Undefined,
                          withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'triangleLeft'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "triangleLeft"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_triangleRight(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                           brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined,
                           dy=Undefined, enableSmoothPath=Undefined, inlineLegend=Undefined,
                           legendTitle=Undefined, linePattern=Undefined, linkConnectionType=Undefined,
                           linkMinHeight=Undefined, linkStyle=Undefined, matrixExtent=Undefined,
                           mouseOver=Undefined, outline=Undefined, outlineWidth=Undefined,
                           select=Undefined, textAnchor=Undefined, textFontSize=Undefined,
                           textFontWeight=Undefined, textStroke=Undefined, textStrokeWidth=Undefined,
                           withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'triangleRight'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "triangleRight"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_triangleBottom(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                            brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined,
                            dy=Undefined, enableSmoothPath=Undefined, inlineLegend=Undefined,
                            legendTitle=Undefined, linePattern=Undefined, linkConnectionType=Undefined,
                            linkMinHeight=Undefined, linkStyle=Undefined, matrixExtent=Undefined,
                            mouseOver=Undefined, outline=Undefined, outlineWidth=Undefined,
                            select=Undefined, textAnchor=Undefined, textFontSize=Undefined,
                            textFontWeight=Undefined, textStroke=Undefined, textStrokeWidth=Undefined,
                            withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'triangleBottom'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "triangleBottom"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_brush(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                   brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                   enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                   linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                   linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                   outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                   textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                   textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'brush'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "brush"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy

    def mark_header(self: T, align=Undefined, background=Undefined, backgroundOpacity=Undefined,
                    brush=Undefined, curve=Undefined, dashed=Undefined, dx=Undefined, dy=Undefined,
                    enableSmoothPath=Undefined, inlineLegend=Undefined, legendTitle=Undefined,
                    linePattern=Undefined, linkConnectionType=Undefined, linkMinHeight=Undefined,
                    linkStyle=Undefined, matrixExtent=Undefined, mouseOver=Undefined, outline=Undefined,
                    outlineWidth=Undefined, select=Undefined, textAnchor=Undefined,
                    textFontSize=Undefined, textFontWeight=Undefined, textStroke=Undefined,
                    textStrokeWidth=Undefined, withinLinkVerticalLines=Undefined, **kwds) -> T:
        """Set the track's mark to 'header'
    
        For information on additional arguments, see :class:`Style`
        """
        kwds = dict(align=align, background=background, backgroundOpacity=backgroundOpacity,
                    brush=brush, curve=curve, dashed=dashed, dx=dx, dy=dy,
                    enableSmoothPath=enableSmoothPath, inlineLegend=inlineLegend,
                    legendTitle=legendTitle, linePattern=linePattern,
                    linkConnectionType=linkConnectionType, linkMinHeight=linkMinHeight,
                    linkStyle=linkStyle, matrixExtent=matrixExtent, mouseOver=mouseOver,
                    outline=outline, outlineWidth=outlineWidth, select=select, textAnchor=textAnchor,
                    textFontSize=textFontSize, textFontWeight=textFontWeight, textStroke=textStroke,
                    textStrokeWidth=textStrokeWidth, withinLinkVerticalLines=withinLinkVerticalLines,
                    **kwds)
        copy = self.copy()
        copy.mark = "header"
        if any(val is not Undefined for val in kwds.values()):
            copy.style = core.Style(**kwds)
        return copy


"""
Text Marks
==========
"""
# category: basic marks
import gosling as gos

data = gos.multivec(
    url="https://resgen.io/api/v1/tileset_info/?d=UvVPeLHuRDiYA3qwFlm7xQ",
    row="base",
    column="position",
    value="count",
    categories=["A", "T", "G", "C"],
    start="start",
    end="end",
    binSize=16,
)

track = gos.Track(data).mark_text(textStrokeWidth=0).encode(
    y="count:Q",
    x=gos.X("start:G", axis="top"),
    xe="end:G",
    color=gos.Color("base:N", domain=["A", "T", "G", "C"]),
    text="base:N",
).properties(width=725, height=180, stretch=True)

track.view(title="Basic Marks: Text", subtitle="Tutorial Examples")



"""
Comparative Matrices
====================
"""
# category: composite visualizations
import gosling as gos

size = 375
scaled_size = size / 15

# Data

HFFc6_H3K4me3 = gos.bigwig(
    url="https://s3.amazonaws.com/gosling-lang.org/data/HFFc6_H3K4me3.bigWig",
    column="position",
    value="peak",
    binSize=8,
)

HFFc6_Atacseq = gos.bigwig(
    url="https://s3.amazonaws.com/gosling-lang.org/data/HFFc6_Atacseq.mRp.clN.bigWig",
    column="position",
    value="peak",
    binSize=8
)

HFFC6_CTCF = gos.bigwig(
    url="https://s3.amazonaws.com/gosling-lang.org/data/HFFC6_CTCF.mRp.clN.bigWig",
    column="position",
    value="peak",
    binSize=8,
)

genes = gos.beddb(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=gene-annotation",
    genomicFields=[
        {"index": 1, "name": "start"},
        {"index": 2, "name": "end"},
    ],
    valueFields=[
        {"index": 5, "name": "strand", "type": "nominal"},
        {"index": 3, "name": "name", "type": "nominal"},
    ],
)

micro_c = gos.matrix("https://server.gosling-lang.org/api/v1/tileset_info/?d=hffc6-microc-hg38")
hi_c = gos.matrix("https://server.gosling-lang.org/api/v1/tileset_info/?d=hffc6-hic-hg38")

epilogos_data = gos.multivec(
    url="https://server.gosling-lang.org/api/v1/tileset_info/?d=epilogos-hg38",
    row="category",
    column="position",
    value="value",
    categories=[
        "Active TSS", "Flanking Active TSS", "Transcr at gene 5\\' and 3\\'",
        "Strong transcription", "Weak transcription", "Genic enhancers",
        "Enhancers", "ZNF genes & repeats", "Heterochromatin",
        "Bivalent/Poised TSS", "Flanking Bivalent TSS/Enh", "Bivalent Enhancer",
        "Repressed PolyComb", "Weak Repressed PolyComb", "Quiescent/Low",
    ],
    binSize=8,
)

# Tracks & Views

track1 = gos.Track(HFFc6_H3K4me3).mark_bar().encode(
    x=gos.X("start:G", axis="top"),
    xe="end:G",
    y=gos.Y("peak:Q", axis="none"),
    color=gos.value("darkgreen"),
).properties(title="HFFc6_H3K4me3", height=size, width=scaled_size)

track2 = gos.Track(HFFc6_Atacseq).mark_bar().encode(
    x="start:G",
    xe="end:G",
    y=gos.Y("peak:Q", axis="none"),
    color=gos.value("#E79F00"),
).properties(title="HFFc6_ATAC", height=size, width=scaled_size)

gene_anno_base = gos.Track(genes).encode(
    x="start:G",
    size=gos.value(13),
    stroke=gos.value("white"),
    strokeWidth=gos.value(1),
    color=gos.value("#CB7AA7"),
    row=gos.Row("strand:N", domain=["+", "-"]),
)

gene_overlay = gos.overlay(
    gos.Track(HFFC6_CTCF).mark_bar().encode(
        x="start:G",
        xe="end:G",
        y=gos.Y("peak:Q", axis="none"),
        color=gos.value("#0072B2")
    ),
    gene_anno_base.mark_triangleRight(backgroundOpacity=0).encode(
        color=gos.value("#CB7AA7"),
    ).transform_filter("strand", oneOf=["+"]),
    gene_anno_base.mark_triangleLeft(backgroundOpacity=0).encode(
        color=gos.value("#029F73"),
    ).transform_filter("strand", oneOf=["-"]).properties(title="HFFC6_CTCF")
).properties(height=size, width=scaled_size)

# Configure Layout

left = gos.stack(track1, track2, gene_overlay).properties(
    orientation="vertical",
    yOffset=size / 2.78,
)

top = gos.stack(
    track1.properties(width=size, height=scaled_size),
    track2.properties(width=size, height=scaled_size),
    gene_overlay.properties(width=size, height=scaled_size),
)

matrix = gos.Track(micro_c).mark_rect().encode(
    x="xs:G",
    xe="xe:G",
    y="ys:G",
    ye="ye:G",
    color=gos.Color("value:Q", range="warm"),
).properties(title="HFFc6_Micro-C", width=size, height=size)


bottom = gos.Track(epilogos_data).mark_bar().encode(
    x=gos.X("start:G", axis="none"),
    xe="end:G",
    y=gos.Y("value:Q", axis="none"),
    color=gos.Color("category:N", range=[
        "#FF0000", "#FF4500", "#32CD32", "#008000", "#006400",
        "#C2E105", "#FFFF00", "#66CDAA", "#8A91D0", "#CD5C5C",
        "#E9967A", "#BDB76B", "#808080", "#C0C0C0", "gray"
    ]),
).transform_filter("value", inRange=[0, 999999]).properties(
    title="Epilogos (hg38)",
    width=size,
    height=scaled_size,
)


left_matrix = gos.vertical(top, matrix, bottom, spacing=0)
right_matrix = gos.vertical(
    top,
    # replace data source for right side, change title
    matrix.properties(data=hi_c, title="HFFc6_Hi-C"),
    bottom,
    spacing=0,
)

right = gos.View(views=[left_matrix, right_matrix], spacing=30)

gos.horizontal(left, right).properties(
    title="Matrix Visualization",
    subtitle="Comparison of Micro-C and Hi-C for HFFc6 Cells",
    xDomain=gos.GenomicDomain(chromosome="chr7", interval=[77700000, 81000000]),
    spacing=1,
    linkingId="-"
)
